﻿
;- Inotify declarations

; http://man7.org/linux/man-pages/man7/inotify.7.html
; http://man7.org/linux/man-pages/man2/read.2.html
; http://man7.org/linux/man-pages/man2/inotify_add_watch.2.html

Structure inotify_event
  wd.l      ; watch descriptor
  mask.l    ; mask describing event
  cookie.l  ; unique cookie associating related events for rename2))
  len.l     ; size of name field
  name.a[0]  ; optional null-terminated name
EndStructure

; Syscall numbers, linux 64-bit
! sys_read = 0
! sys_close = 3
! sys_inotify_init = 253
! sys_inotify_add_watch = 254
! sys_inotify_rm_watch = 255

; Supported events suitable For MASK parameter of INOTIFY_ADD_WATCH.
! IN_ACCESS = $00000001 ; File was accessed.
! IN_MODIFY = $00000002 ; File was modified.
! IN_ATTRIB = $00000004 ; Metadata changed.
! IN_CLOSE_WRITE   = $00000008   ; Writable file was closed.
! IN_CLOSE_NOWRITE =  $00000010  ; Unwrittable file closed.
! IN_CLOSE = IN_CLOSE_WRITE or IN_CLOSE_NOWRITE ; Close.
! IN_OPEN        = $00000020   ; File was opened.
! IN_MOVED_FROM  = $00000040   ; File was moved from X.
! IN_MOVED_TO    =  $00000080  ; File was moved To Y.
! IN_MOVE        = IN_MOVED_FROM or IN_MOVED_TO ; Moves.
! IN_CREATE      = $00000100   ; Subfile was created.
! IN_DELETE      = $00000200   ; Subfile was deleted.
! IN_DELETE_SELF = $00000400   ; Self was deleted.
! IN_MOVE_SELF   = $00000800   ; Self was moved.
! IN_ALL_EVENTS	= IN_ACCESS or IN_MODIFY or IN_ATTRIB or IN_CLOSE_WRITE Or IN_CLOSE_NOWRITE Or IN_OPEN Or IN_MOVED_FROM Or IN_MOVED_TO Or IN_CREATE Or IN_DELETE Or IN_DELETE_SELF Or IN_MOVE_SELF
; The same in PB:
#IN_ACCESS = $00000001 ; File was accessed.
#IN_MODIFY = $00000002 ; File was modified.
#IN_ATTRIB = $00000004 ; Metadata changed.
#IN_CLOSE_WRITE   = $00000008   ; Writable file was closed.
#IN_CLOSE_NOWRITE =  $00000010  ; Unwrittable file closed.
#IN_CLOSE = #IN_CLOSE_WRITE | #IN_CLOSE_NOWRITE ; Close.
#IN_OPEN        = $00000020   ; File was opened.
#IN_MOVED_FROM  = $00000040   ; File was moved from X.
#IN_MOVED_TO    =  $00000080  ; File was moved To Y.
#IN_MOVE        = #IN_MOVED_FROM | #IN_MOVED_TO ; Moves.
#IN_CREATE      = $00000100   ; Subfile was created.
#IN_DELETE      = $00000200   ; Subfile was deleted.
#IN_DELETE_SELF = $00000400   ; Self was deleted.
#IN_MOVE_SELF   = $00000800   ; Self was moved.

Procedure.s IN_EventMaskToString(Mask)
  Protected Result.s
  If Mask & #IN_ACCESS:        Result + "IN_ACCESS " : EndIf
  If Mask & #IN_ATTRIB:        Result + "IN_ATTRIB " : EndIf
  If Mask & #IN_CLOSE_NOWRITE: Result + "IN_CLOSE_NOWRITE " : EndIf
  If Mask & #IN_CLOSE_WRITE:   Result + "IN_CLOSE_WRITE " : EndIf
  If Mask & #IN_CREATE:        Result + "IN_CREATE " : EndIf
  If Mask & #IN_DELETE:        Result + "IN_DELETE " : EndIf
  If Mask & #IN_DELETE_SELF:   Result + "IN_DELETE_SELF " : EndIf
  ;If Mask & #IN_IGNORED:       Result + "IN_IGNORED " : EndIf
  ;If Mask & #IN_ISDIR:         Result + "IN_ISDIR " : EndIf
  If Mask & #IN_MODIFY:        Result + "IN_MODIFY " : EndIf
  If Mask & #IN_MOVE_SELF:     Result + "IN_MOVE_SELF " : EndIf
  If Mask & #IN_MOVED_FROM:    Result + "IN_MOVED_FROM " : EndIf
  If Mask & #IN_MOVED_TO:      Result + "IN_MOVED_TO " : EndIf
  If Mask & #IN_OPEN:          Result + "IN_OPEN " : EndIf
  ;If Mask & #IN_Q_OVERFLOW:    Result + "IN_Q_OVERFLOW " : EndIf
  ;If Mask & #IN_UNMOUNT:       Result + "IN_UNMOUNT " : EndIf
  ProcedureReturn Result
EndProcedure

;- Global variables
Global QuitThread = 0
Global DropTargetPath.s
OpenPreferences("maffdroptarget.ini")
  DropTargetPath = ReadPreferenceString("DropTargetPath", GetHomeDirectory()+"Maff Drop Target/")
  WritePreferenceString("DropTargetPath", DropTargetPath)
ClosePreferences()
FlushPreferenceBuffers()
If FileSize(DropTargetPath) = -1
  CreateDirectory(DropTargetPath)
EndIf
; Global QueueMutex.i
; Global NewList NotifiedFilesQueue.s()

;- Inotify thread

; All our custom events for communication between the inotify thread and the gui thread
Enumeration #PB_Event_FirstCustomValue
  #InotifyReceiveMessage
EndEnumeration

Procedure sys_read(fd.i, *buf, count.i)
  ! mov rax, sys_read
  ! mov rdi, [p.v_fd]
  ! mov rsi, [p.p_buf]
  ! mov rdx, [p.v_count]
  ! syscall
  ProcedureReturn 
EndProcedure

Procedure.s DQuote(String.s)
  ProcedureReturn #DQUOTE$ + String + #DQUOTE$
EndProcedure

Procedure CreateMaff(Filename.s)
  Delay(2000)
  RunProgram(GetPathPart(ProgramFilename())+"maffsavegui", DQuote(DropTargetPath + Filename), GetPathPart(Filename))
EndProcedure

Procedure HandleINotifyEvent(*event.inotify_event)
  ;Debug "wd: " + Str(*event\wd)
  ;Debug "Mask: " + IN_EventMaskToString(*event\mask)
  Filename.s = PeekS(@*event\name[0], *event\len, #PB_UTF8)
  ;Debug #DOUBLEQUOTE$ + PeekS(@*event\name[0], *event\len, #PB_UTF8) + #DOUBLEQUOTE$
  CreateMaff(Filename)
  ;PostEvent(#InotifyReceiveMessage)
EndProcedure

Procedure ReadINStructsFromBuffer(*buf.inotify_event, bytes_read.i)
;   ShowVariableViewer()
;   ShowMemoryViewer(*buf, bytes_read)
;   CallDebugger
  ;Debug PeekS(@*buf\name[0], *buf\len, #PB_UTF8)
  ;ProcedureReturn 0
  *event.inotify_event = *buf
  ;Debug "bytes_read: " + Str(bytes_read)
  While *event < *buf + bytes_read; - SizeOf(*event) + 1
    HandleINotifyEvent(*event)
    ;Debug *event\len
    *event = *event + SizeOf(inotify_event) + *event\len
  Wend
;   Debug "^^^"
;   Debug *buf
;   Debug *event
;   Debug *event-*buf
;   Debug "---"
  ;Debug PeekS(@*event\name[0], bytes_read, #PB_UTF8)
EndProcedure

Procedure Thread(Value)
  Global fd ; inotify file descriptor
  Protected wd ; watch descriptor
  #IN_READ_BUF_SIZE = 1024*4 + SizeOf(inotify_event)
  Protected *buf.inotify_event = AllocateMemory(#IN_READ_BUF_SIZE)
  ;Protected *event.inotify_event
  Protected *Pathname = AllocateMemory(4096)
  PokeS(*Pathname, DropTargetPath, 4000, #PB_UTF8)
  ! mov rax, sys_inotify_init
  ! syscall
  ! mov [v_fd], rax
  
  ;Debug fd
  
  ! mov rax, sys_inotify_add_watch
  ! mov rdi, [v_fd]
  ! mov rsi, [p.p_Pathname]
  ! mov rdx, IN_CLOSE_WRITE
  ! syscall
  ! mov [p.v_wd], rax
  
  ;Debug wd
  
  Repeat 
    bytes_read = sys_read(fd, *buf, #IN_READ_BUF_SIZE)
    If bytes_read >= SizeOf(inotify_event)
      ;HandleINotifyEvent(*buf)
      ;Debug bytes_read
      ;Debug IN_EventMaskToString(*buf\mask)
      ;Debug PeekS(@*buf\name[0], bytes_read, #PB_UTF8)
      ReadINStructsFromBuffer(*buf, bytes_read)
    Else
      If #PB_MessageRequester_No = MessageRequester("Maff Droptarget", "Invalid inotify request. Keep trying?", #PB_MessageRequester_YesNo)
;         If #PB_MessageRequester_Yes
;           RunProgram(
;         EndIf
        End
      EndIf
      ;CallDebugger
      ;End
    EndIf
  ForEver ; bytes_read < 1 Or QuitThread = 1
EndProcedure

;- Main

; OpenWindow(0, 200, 200, 100, 100, "Maff Droptarget");, #PB_Window_Invisible)
; 
; Thread = CreateThread(@Thread(), 0)
Thread(0)
; ;End
; 
; Repeat
;   Event = WaitWindowEvent()
;   
;   Select Event
;     Case #InotifyReceiveMessage
;       ;Debug "Thread begin processing "
;       
;   EndSelect
;   
; Until Event = #PB_Event_CloseWindow
; QuitThread = 1
; 
; WaitThread(Thread, 60)
; KillThread(Thread)

; Close inotify file descriptor
! mov rax, sys_close
! mov rdi, [v_fd]
! syscall





; IDE Options = PureBasic 5.60 (Linux - x64)
; ExecutableFormat = Console
; Folding = --
; EnableXP
; Executable = release-linux/maffdroptarget
; CompileSourceDirectory