﻿
; Create a MAFF archive

; Takes an index file, and assumes the corresponding files are placed in a folder
; called indexfile_files

; Log file
;CreateFile(0, "/home/trond/Dropbox/drop-src/maff-native/maffcreate.log")

EnableExplicit

CompilerIf #PB_Compiler_OS = #PB_OS_Windows
  #PATHSLASH$ = "\"
CompilerElse
  #PATHSLASH$ = "/"
CompilerEndIf

Global InIndexFileName.s = ""
Global InFolder.s = ""
Global OutFileName.s = ""
Global MaffDataFolder.s = ""
Global FilesFolder.s = ""
Global NewMap OldNewFilenames.s(2048)

Procedure Abort(Error.s)
  PrintN("MAFFCreate: Fatal error: " + Error)
  CompilerIf #PB_Compiler_Debugger
    Input()
  CompilerEndIf
  End -1
EndProcedure

Procedure GetProgramParameters()
  Protected C = CountProgramParameters()
  If C = 2
    InIndexFileName = ProgramParameter(0)
    OutFileName = ProgramParameter(1)
    ;PrintN("InIndexFileName: " + InIndexFileName)
  EndIf
  If C <> 2 Or InIndexFileName= "" Or OutFileName = ""
    Protected Msg.s = "Invalid number of parameters." + #CRLF$
    Msg + "Usage: maffcreate input_index_file output_file" + #CRLF$
    Msg + "Possible solution: Put quotes around parameters with spaces in them."
    Abort(Msg)
  EndIf
EndProcedure

Procedure PackfileCallback(FolderName.s, FN.s)
  ; WriteString(0, FolderName+FN, #PB_UTF8)
  AddPackFile(0, FolderName+FN, MaffDataFolder + FilesFolder + FN)
EndProcedure

; Procedure DeleteFileCallback(FolderName.s, FileName.s)
;   DeleteFile(FolderName + FileName)
; EndProcedure

Prototype ExamineDirectoryCallback(FolderName.s, FileName.s)

Procedure GetFilenames(FolderName.s, FileCallback.ExamineDirectoryCallback, DirCbk.ExamineDirectoryCallback = 0)
  Protected Directory = ExamineDirectory(#PB_Any, FolderName, "*")
  Protected FN.s
  If Directory
    If Right(FolderName, 1) <> #PATHSLASH$
      FolderName + #PATHSLASH$
    EndIf
    While NextDirectoryEntry(Directory)
      FN.s = DirectoryEntryName(Directory)
      If DirectoryEntryType(Directory) = #PB_DirectoryEntry_Directory
        If FN = "." Or FN = ".."
          Continue
        EndIf
        GetFilenames(FolderName+FN, FileCallback)
        If DirCbk
          DirCbk(FolderName, FN)
        EndIf
      Else
        ; Application code
        FileCallback(FolderName, FN)
        ; End application code
      EndIf
    Wend
    FinishDirectory(Directory)
  Else
    Abort("Couldn't read from directory: " + FolderName)
  EndIf
EndProcedure

; Main:
UseZipPacker()
OpenConsole("MAFFCreate")
GetProgramParameters()

InFolder = GetPathPart(InIndexFileName) ;ReverseString(Mid(ReverseString(InIndexFileName), FindString(ReverseString(InIndexFileName), #PATHSLASH$)))
MaffDataFolder.s = FormatDate("%yyyy-%mm-%dd_%hh.%ii.%ss", Date()) + #PATHSLASH$

Define FilesFolderPrefix.s = GetFilePart(InIndexFileName, #PB_FileSystem_NoExtension)
PrintN(FilesFolderPrefix)
If ExamineDirectory(0, InFolder, FilesFolderPrefix+"*")
  While NextDirectoryEntry(0)
    If DirectoryEntryType(0) = #PB_DirectoryEntry_Directory
      FilesFolder.s = DirectoryEntryName(0)
    EndIf
  Wend
  FinishDirectory(0)
EndIf
PrintN(FilesFolder)
If FilesFolder = ""
  Abort("Files folder not found")
EndIf
FilesFolder = FilesFolderPrefix + "_files"

Procedure.s UniqueFilename(OutFileName.s)
  Protected NewFileName.s = OutFileName
  Protected I = 1
  Protected Extension.s
  While FileSize(NewFileName) <> -1
    I + 1
    NewFileName = GetPathPart(OutFileName) + GetFilePart(OutFileName, #PB_FileSystem_NoExtension) + "_(" + Str(I) + ")"
    Extension.s = GetExtensionPart(OutFileName)
    If Extension
    
      NewFileName + "." + Extension
    EndIf
  Wend
  If OutFileName <> NewFileName
    PrintN("Output file exists, using another filename: " + NewFileName)
    OutFileName = NewFileName
  EndIf
  ProcedureReturn OutFileName
EndProcedure

; Don't overwrite files - add a number instead
; CheckedOutFileName.s = OutFileName
; I = 0
; While FileSize(CheckedOutFileName) <> -1
;   I + 1
;   CheckedOutFileName = GetPathPart(OutFileName) + GetFilePart(OutFileName, #PB_FileSystem_NoExtension) + "_" + Str(I) + "." + GetExtensionPart(OutFileName)
; Wend
; If OutFileName <> CheckedOutFileName
;   PrintN("Output file exists, using another filename: " + CheckedOutFileName)
;   OutFileName = CheckedOutFileName
; EndIf
OutFileName = UniqueFilename(OutFileName)

If CreatePack(0, OutFileName, #PB_PackerPlugin_Zip, 8)
  If Not AddPackFile(0, InIndexFileName, MaffDataFolder + "index.html")
    Abort("Couldn't pack index file")
  EndIf
  
;   PrintN("InIndexFileName: " + InIndexFileName)
;   PrintN("InFolder: " + InFolder)
;   PrintN("FilesFolder: " + FilesFolder)
  
  GetFilenames(InFolder + FilesFolder, @PackfileCallback())
  ClosePack(0)
  DeleteFile(InIndexFileName)
  PrintN("Deleting "+ InFolder+FilesFolder)
  DeleteDirectory(InFolder + FilesFolder, "", #PB_FileSystem_Recursive | #PB_FileSystem_Force)
Else
  Abort("Couldn't create output file: " + OutFileName)
EndIf
; IDE Options = PureBasic 5.60 (Linux - x64)
; ExecutableFormat = Console
; Folding = --
; EnableXP
; Executable = release-linux/maffcreate
; CommandLine = "/home/trond/Dropbox/drop-src/maff-native/osnews_example/slim - Google-søk.html" "/home/trond/Dropbox/drop-src/maff-native/output/slim.maff"