﻿
; MAFF Server

; Valid URLs
; localhost:port/read/localfilepath
; localhost:port/identify/

Procedure Abort(Error.s)
  MessageRequester("MAFF","MAFFServe: Fatal error: " + Error)
  End -1
EndProcedure

; Mime types
Global NewMap Mime.s()
Mime("css") = "text/css"
Mime("es") = "application/ecmascript"
Mime("gif") = "image/gif"
Mime("htm") = "text/html"
Mime("html") = "text/html"
Mime("ico") = "image/x-icon"
Mime("jpeg") = "image/jpeg"
Mime("jpg") = "image/jpeg"
Mime("js") = "application/javascript"
Mime("json") = "application/json"
Mime("oga") = "audio/ogg"
Mime("ogv") = "video/ogg"
Mime("ogx") = "application/ogg"
Mime("otf") = "font/otf"
Mime("png") = "image/png"
Mime("pdf") = "application/pdf"
Mime("svg") = "image/svg+xml"
Mime("tif") = "image/tiff"
Mime("tiff") = "image/tiff"
Mime("ts") = "application/typescript"
Mime("ttf") = "font/ttf"
Mime("wav") = "audio/wav"
Mime("weba") = "audio/webm"
Mime("webm") = "video/webm"
Mime("webp") = "image/webp"
Mime("woff") = "font/woff"
Mime("woff2") = "font/woff2"
Mime("xhtml") = "application/xhtml+xml"
Mime("xml") = "application/xml"
Mime("mpeg") = "video/mpeg"
;Mime("") = ""

; HTTP headers
Structure SHTTPResponse
  Status.s
  Map Headers.s()
  Contents.s
  ContentsBinaryLength.i
  *ContentsBinary
EndStructure

Procedure SendHTTPResponse(ClientID, *R.SHTTPResponse)
  Response.s = *R\Status + #CRLF$
  ForEach *R\Headers()
    Response + MapKey(*R\Headers()) + ": " + *R\Headers() + #CRLF$
  Next
  Response + "Content-Length:	" + Str(StringByteLength(*r\Contents, #PB_UTF8)) + *r\ContentsBinaryLength + #CRLF$
  Response + #CRLF$
  Response + *R\Contents
  ;Debug Response
  SendNetworkString(ClientID, Response, #PB_UTF8)
  If *R\ContentsBinary
    SendNetworkData(ClientID, *R\ContentsBinary, *r\ContentsBinaryLength)
  EndIf
  CloseNetworkConnection(ClientID)
EndProcedure

Procedure FillDefaultHTTPResponse(*R.SHTTPResponse, Contents.s = "", Status.s = "HTTP/1.1 200 OK")
  ;ResetStructure(*R, SHTTPResponse)
  *R\Status = Status
  *R\Contents = Contents
  *R\ContentsBinaryLength = 0
  *R\ContentsBinary = 0
  *R\Headers("Connection") = "close"
  *R\Headers("Server") = "MaffServe"
  *R\Headers("Content-Type") = "text/html; charset=utf-8"
  *R\Headers("Cache-Control") = "no-cache"
  ProcedureReturn *R
EndProcedure

Procedure NewDefaultHTTPResponse(Contents.s = "", Status.s = "HTTP/1.1 200 OK")
  *R.SHTTPResponse = AllocateStructure(SHTTPResponse)
  ProcedureReturn FillDefaultHTTPResponse(*R, Contents, Status)
EndProcedure

;- Init
UseZipPacker()

If InitNetwork() = 0
  Abort("Can't initialize the network !")
  End
EndIf


Define Port
Define DelayMinTime.d
Define DelayMaxTime.d

Define PrefFile.s = GetPathPart(ProgramFilename())+"maff.ini"
OpenPreferences(PrefFile)
  PreferenceGroup("Server")
  Port = ReadPreferenceInteger("Port", 8082)
  WritePreferenceInteger("Port", Port)
  DelayMinTime = ReadPreferenceDouble("DelayMinTime", 2)
  WritePreferenceDouble("DelayMinTime", DelayMinTime)
  DelayMaxTime = ReadPreferenceDouble("DelayMaxTime", 150)
  WritePreferenceDouble("DelayMaxTime", DelayMaxTime)
ClosePreferences()
FlushPreferenceBuffers()

Global NewMap ClientData.s()
#RcvSize = 65536
Define *RcvBuffer = AllocateMemory(#RcvSize)

Procedure OpenPackFailed(ClientID, *R.SHTTPResponse, MaffPath.s)
  If FileSize(MaffPath) = -1
    FillDefaultHTTPResponse(*R, "404 Not Found. MAFF file not found.", "HTTP/1.1 404 Not Found")
  Else
    FillDefaultHTTPResponse(*R, "503 Service Unavailable. MAFF file in use by another process, or corrupt.", "HTTP/1.1 503 Service Unavailable")
  EndIf
EndProcedure

Procedure RedirectToIndexFile(ClientID, MaffPath.s)
  ;Debug "MaffPath: " + MaffPath
  If OpenPack(0, MaffPath)
    If ExaminePack(0)
      While NextPackEntry(0)
        Name.s = PackEntryName(0)
        If CountString(Name, "/") = 1
          FN.s = StringField(Name, 2, "/")
          If FN = "index.html" Or FN = "index.htm" Or FN = "index.xhtml"
            FillDefaultHTTPResponse(R.SHTTPResponse, "302 Found", "HTTP/1.1 302 Found")
            R\Headers("Location") = Name
            SendHTTPResponse(ClientID, R)
            ProcedureReturn 1 ; sic
          EndIf
        EndIf
      Wend
    EndIf
  EndIf
  ProcedureReturn 0
EndProcedure

Procedure ReturnReadRequest(ClientID.i, ReadPath.s)
  If Right(ReadPath, 5) = ".maff"
    FillDefaultHTTPResponse(R.SHTTPResponse, "302 Found", "HTTP/1.1 302 Found")
    R\Headers("Location") = "/read/" + ReadPath + "/"
    SendHTTPResponse(ClientID, R)
    ProcedureReturn 1
  EndIf
  MaffPathLocal.s = StringField(ReadPath, 1, ".maff/") + ".maff"
  QueryString.s = StringField(ReadPath, 2, ".maff/")
  If QueryString = ""
    If RedirectToIndexFile(ClientID, MaffPathLocal)
      ProcedureReturn 2
    EndIf
  EndIf
  ; Read maff archive
  If Not OpenPack(0, MaffPathLocal)
    OpenPackFailed(ClientID, R.SHTTPResponse, MaffPathLocal)
  Else
    If Not ExaminePack(0)
      FillDefaultHTTPResponse(R, "500 Internal Server Error. Can't unpack maff file.", "HTTP/1.1 500 Internal Server Error")
    Else
      FoundFile = 0
      While NextPackEntry(0)
        If PackEntryName(0) = QueryString
          FillDefaultHTTPResponse(R.SHTTPResponse)
          R\ContentsBinaryLength = PackEntrySize(0)
          R\ContentsBinary = AllocateMemory(R\ContentsBinaryLength)
          If UncompressPackMemory(0, R\ContentsBinary, R\ContentsBinaryLength) <> R\ContentsBinaryLength
            FreeMemory(R\ContentsBinary)
            FillDefaultHTTPResponse(R2.SHTTPResponse, "500 Internal Server Error. Can't unpack maff file.", "HTTP/1.1 500 Internal Server Error")
            SendHTTPResponse(ClientID, R2)
            ClosePack(0)
            ProcedureReturn 0
          EndIf
          R\Headers("Content-Type") = Mime(GetExtensionPart(QueryString))
          FoundFile = 1
          Break
        EndIf
      Wend
      If FoundFile = 0
        FillDefaultHTTPResponse(R, "404 Not Found. Resource in MAFF file not found.", "HTTP/1.1 404 Not Found.")
      EndIf
    EndIf
    ClosePack(0)
  EndIf
  
  SendHTTPResponse(ClientID, R)
  If R\ContentsBinary
    FreeMemory(R\ContentsBinary)
  EndIf
EndProcedure

Procedure HandleWgetRequest(ClientID, ReadPath.s)
  FillDefaultHTTPResponse(R.SHTTPResponse, "Received wget request - unimplemented")
  SendHTTPResponse(ClientID, R)
EndProcedure

Procedure HandleRequest(ClientID)
  cd.s = ClientData(Str(ClientID))
  RequestPath.s = StringField(cd, 2, " ")
  Action.s = StringField(RequestPath, 2, "/")
  R.SHTTPResponse
  Select Action
    Case "read"
      ReadPath.s = Mid(URLDecoder(RequestPath), Len("/read/")+1)
      ReturnReadRequest(ClientID, ReadPath.s)
    Case "identify"
      FillDefaultHTTPResponse(R, "maffserve")
      SendHTTPResponse(ClientID, R)
    Case "wget"
      HandleWgetRequest(ClientID, ReadPath)
    Default
      FillDefaultHTTPResponse(R, "400 Bad Request.", "HTTP/1.1 400 Bad Request")
      SendHTTPResponse(ClientID, R)
  EndSelect
EndProcedure


H = 350
W = 500

;OpenWindow(0, 200, 300, W, H, "MaffServe", #PB_Window_MaximizeGadget | #PB_Window_SizeGadget)

If CreateNetworkServer(0, Port)
  
  Repeat
    SEvent = NetworkServerEvent()
    If SEvent
      DelayTime.d = DelayMinTime
      
      ClientID = EventClient()
      
      Select SEvent
       
        Case #PB_NetworkEvent_Connect
          ClientData(Str(ClientID)) = ""
          If GetClientIP(ClientID) <> MakeIPAddress(127,0,0,1)
            CloseNetworkConnection(ClientID)
          EndIf
          
        Case #PB_NetworkEvent_Data
          ReceiveNetworkData(ClientID, *RcvBuffer, #RcvSize)
          RcvStr.s = PeekS(*RcvBuffer, #RcvSize, #PB_UTF8)
          ClientData(Str(ClientID)) + RcvStr
          If FindString(RcvStr, #CRLF$+#CRLF$)
            HandleRequest(ClientID)
          EndIf
        Case #PB_NetworkEvent_Disconnect
          DeleteMapElement(ClientData(), Str(ClientID))
    
      EndSelect
      ; Event = WindowEvent()
    Else
      If DelayTime < DelayMaxTime
        DelayTime + 0.1
        Debug DelayTime
      EndIf
      Delay(DelayTime)
      ;Event = WaitWindowEvent(DelayTime)
    EndIf
    
;     If Event = #PB_Event_CloseWindow
;       Quit = 1
;     EndIf
    
  Until Quit = 1 
  
  CloseNetworkServer(0)
Else
  Abort("Can't create the server (port in use?). " + Str(Port))
EndIf

End
; IDE Options = PureBasic 5.60 (Linux - x64)
; Folding = --
; EnableXP
; Executable = release-linux/maffserve
; CompileSourceDirectory
; EnablePurifier